local bzUtils = require "mCore";

local mFlow = bzUtils.core.mModule:new("mFlow");
bzUtils:addModule(mFlow);

local callback = {
    new = function(cls,name,func,env)
        local cback = setmetatable({},cls:mt());
        cback.name = name;
        cback.func = func;
        mFlow:addCallback(cback,env);
        return cback;
    end,
    prototype = {},
    mt = function(cls)
        return {
            __index = cls.prototype,
            __call = cls.new
        }; 
    end
};
setmetatable(callback,callback:mt());
local event = {
    new = function(cls,name,...)
        local evt = setmetatable({},cls:mt());
        evt.name = name;
        evt.args = {...};
        return evt;
    end,
    prototype = {},
    mt = function(cls)
        return {
            __index = cls.prototype,
            __call = cls.new
        }; 
    end 
};

setmetatable(event,event:mt());
local eventHandler = {
    new = function(cls,parent,def)
        local listener = setmetatable(def or {},cls:mt());
        listener.parent = parent or listener; 
        listener.eListeners = {};
        return listener;
    end,
    Load = function(cls,data,p)
        print("EventListener load");
        self = cls:new(p,data);
        self.eListeners = data["listeners"];
        --[[for i2,v2 in pairs(data["listeners"]) do
            for i,v in pairs(v2) do
                if(v.f) then
                    v.name = loadstring(v.name);
                end
            end
        end--]]
        data["listeners"] = nil;
        return self;
    end,
    prototype = {
        saveable = true,
        addListener = function(self,mEventName,mCallback,env)
            local c_name = "";
            local isFunc = false;
            print(mCallback,type(mCallback));
            if(type(mCallback) == "function") then
                c_name = mCallback;
                isFunc = true;
            else
                if(type(mCallback) == "table") then
                    c_name = mCallback.name;
                else
                    if(not mFlow:getCallback(mCallback)) then
                        error("No callback named " .. tostring(mCallback) .. " exists!",3);
                    end
                    c_name = mCallback;
                end
               
            end
            if(not self.eListeners[mEventName]) then
                self.eListeners[mEventName] = {};
            end
            table.insert(self.eListeners[mEventName],{f=isFunc,name=c_name,env=env});
        end,
        Save = function(self)
            local l = {};
            for i2,v2 in pairs(self.eListeners) do
                l[i2] = {};
                for i,v in pairs(v2) do
                    local a = {name = v.name,f = v.f,env=v.env};
                    if(a.f) then
                        a.name = string.dump(a.name);
                    end
                    table.insert(l[i2],a);
                end
            end
            if not self._currentSave then
                self._currentSave = {listeners = l};
            end
            return self._currentSave;
        end,
        SaveDone = function(self)
            self._currentSave = nil;
        end,
        dispatchEvent = function(self,mEvent)
            --return stack
            local ret = {};
            if(self.eListeners[mEvent.name]) then
                for i,v in pairs(self.eListeners[mEvent.name]) do
                    local f = nil
                    if(v.f) then
                        f = v.name;
                    else
                        f = mFlow:getCallback(v.name,v.env).func;
                    end
                    setfenv(f,setmetatable({self = self.parent,this = self.parent},{__index = _G,__newindex = _G}));
                    table.insert(ret,{f(self.parent,unpack(mEvent.args))});
                end
            end
            return ret;
        end
    },
    mt = function(cls)
        return {
            __index = cls.prototype,
            __call = cls.new
        }; 
    end

};
setmetatable(eventHandler,eventHandler:mt());

mFlow.callbacks = {};

mFlow.addCallback = function(self,callback,env)
    local a = env or self.callbacks;
    if(not a[callback.name]) then
        a[callback.name] = callback;
    else
        error(("Callback %s already exist!"):format(callback.name),2);
    end
end
mFlow.getCallback = function(self,name,evn)
    return (env and env[name]) or self.callbacks[name];
end

mFlow.callback = callback;
mFlow.event = event;
mFlow.eventHandler = eventHandler;

mFlow.Save = function(self)
    local mt = getmetatable(self);
    r = mt.__index.Save(self);
    r[0] = {};
    return r;
end

mFlow.Load = function(self,data)
    local mt = getmetatable(self);
    local c = data[0];
    
    
    mt.__index.Load(self,data);
end

bzUtils.mFlow = mFlow;

return bzUtils.mFlow;