--[[
    Module for creating object wrappers for handles
]]
local bzUtils = require "mCore";
local mFlow = require "mFlow";
local mHandle = bzUtils.core.mModule("mHandle");
bzUtils:addModule(mHandle);
--Table with handle object pairs
mHandle.objects = {};

--[[mHandle.getObject = function(self,handle)
    if(not self.objects[handle]) then
        self.objects[handle] = self.gameObject(handle);
    end
    return self.objects[handle];
end--]]
local gameObject = nil;
gameObject = {
    new = function(cls,handle,def)
        if(mHandle.objects[handle]) then
            return mHandle.objects[handle];
        end
        local inst = setmetatable(def or {},cls.mt);
        inst.h = handle;
        if(not def) then
            inst.eventHandler = mFlow.eventHandler(inst);
            inst.lastTarget = nil;
            inst.dead = false;
            for i,v in pairs(inst.genericGetMap) do
                inst:genericGetFunc(v[1],v[2]);
            end
        end
        inst.old = {};
        local r = bzUtils.core.createLocalStore(inst);
        mHandle.objects[handle] = r;
        return r;
    end,
    Load = function(cls,data)
        local cData = {};
        for i,v in pairs(data) do
            cData[i] = v;
            data[i] = nil;
        end
        data.eventHandler = mFlow.eventHandler:Load(cData.eHandler,data);
        data.dead = cData.dead;
        local inst = cls:new(cData.h,data);
        local uStore = cData["user_store"];
        cData["eHandler"] = nil;
        cData["user_store"] = nil;
        cData["dead"] = nil;
        for i,v in pairs(cData) do
            rawset(inst,i,v);
            inst.old[i] = v;
        end
        for i,v in pairs(uStore) do
            inst[i] = v;
        end
        return inst;
    end,
    prototype = {
        saveable = true,
        send = function(self,target,...)
            if(target.message) then
                target:message(self,...);
            end
        end,
        message = function(self,from,...)
            self.eventHandler:dispatchEvent(mFlow.event("receive",from,...));
        end,
        Save = function(self)
            if(not rawget(self,"_currentSave")) then
                local r = {h = self.h,eHandler = self.eventHandler:Save(),lastTarget = self.lastTarget,dead=self.dead};
                for i,v in pairs(self.genericGetMap) do
                    r[v[1]] = rawget(self,v[1]);
                end
                rawset(self,"_currentSave",r);
                bzUtils.core.createSaveTable(self);
            end
            return rawget(self,"_currentSave");
        end,
        SaveDone = function(self)
            rawset(self,"_currentSave",nil);
        end,
        Update = function(self,dtime)
            self.eventHandler:dispatchEvent(mFlow.event("update",dtime));
            --[[if(self.lastTarget ~= self:getTarget()) then
                self.eventHandler:dispatchEvent(mFlow.event("target_changed",self:getTarget(),self.lastTarget));
            end
            --rawset(self,"lastTarget",self:getTarget());
            --]]
            local o = self.old;
            self.old = {};
            for i,v in pairs(self.genericGetMap) do
                local r = self:genericGetFunc(v[1],v[2]);
                self.old[v[1]] = r;
            end
            for i,v in pairs(self.varMonitor) do
                if(o[v] ~= self.old[v]) then
                    self.eventHandler:dispatchEvent(mFlow.event(v .. "_changed",self.old[v],o[v]));
                end
            end
        end,
        DeleteObject = function(self)
            self.eventHandler:dispatchEvent(mFlow.event("delete",dtime));
            self.dead = true;
        end,
        addListener = function(self,mEventName,mCallback,...)
            self.eventHandler:addListener(mEventName,mCallback,...);
        end,    
        dispatchEvent = function(self,mEvent)
            self.eventHandler:dispatchEvent(mEvent);
        end,
        genericGetMap = {
            --var, func_ref
            getMaxHealth = {"maxHealth","GetMaxHealth"},
            getMaxAmmo = {"maxAmmo","GetMaxAmmo"},
            getPilotClass = {"pilotClass","GetPilotClass"},
            getPosition = {"position","GetPosition"},
            getVelocity = {"velocity","GetVelocity"},
            getTeamNum = {"team","GetTeamNum"},
            getTransform = {"transform","GetTransform"},
            getOmega = {"omega","GetOmega"},
            getFront = {"facing","GetFront"},
            getTarget = {"target","GetTarget"},
            getWhoShotMe = {"lastShotBy","GetWhoShotMe"},
            getLastEnemyShot = {"lastEnemyShot","GetLastEnemyShot"},
            getLastFriendShot = {"lastFriendShot","GetLastFriendShot"},
            isDeployed = {"deployed","IsDeployed"},
            isSelected = {"selected","IsSelected"},
            --isCritical = {"critical","IsCritical"},
            getOwner = {"owner","GetOwner"},
            getNation = {"nation","GetNation"},
            getClassId = {"classId","GetClassId"},
            getClassLabel = {"classLabel","GetClassLabel"},
            getClassSig = {"classSig","GetClassSig"},
            getBase = {"base","GetBase"},
            getOdf = {"odf","GetOdf"},
            getLabel = {"label","GetLabel"},
            getCurrentCommand = {"command","GetCurrentCommand"},
            getCurrentWho = {"who","GetCurrentWho"}
        },
        varMonitor = {"team","target","command","pilotClass","owner","who"},
        genericMap = {
            getHealth = "GetHealth",
            getCurHealth = "GetCurHealth",
            setCurHealth = "SetCurHealth",
            getAmmo = "GetAmmo",
            addHealth = "AddHealth",
            addAmmo = "AddAmmo",
            getCurAmmo = "GetCurAmmo",
            setCurAmmo = "SetCurAmmo",
            getNearestObject = "GetNearestObject",
            getNearestVehicle = "GetNearestVehicle",
            getNearestEnemy = "GetNearestEnemy",
            getNearestFriend = "GetNearestFriend",
            getNearestObject = "GetNearestObject",
            getNearestVehicle = "GetNearestVehicle",
            getNearestBuilding = "GetNearestBuilding",
            countUnitsNearObject = "CountUnitsNearObject",
            isWithin = "IsWithin",
            getDistance = "GetDistance",
            deploy = "Deploy",
            setWeaponMask = "SetWeaponMask",
            getWeaponMaks = "GetWeaponMaks",
            getWeaponClass = "GetWeaponClass",
            giveWeapon = "GiveWeapon",
            fireAt = "FireAt",
            damage = "Damage",
            canCommand = "CanCommand",
            canBuild = "CanBuild",
            isBusy = "IsBusy",
            getIndependence = "GetIndependence",
            setIndependence = "SetIndependence",
            setCommand = "SetCommand",
            attack = "Attack",
            goto = "Goto",
            mine = "Mine",
            follow = "Follow",
            defend = "Defend",
            defend2 = "Defend2",
            stop = "Stop",
            patrol = "Patrol",
            retreat = "Retreat",
            getIn = "GetIn",
            pickup = "Pickup",
            dropoff = "Dropoff",
            build = "Build",
            buildAt = "BuildAt",
            hasCargo = "HasCargo",
            getTug = "GetTug",
            ejectPilot = "EjectPilot",
            hopOut = "HopOut",
            killPilot = "KillPilot",
            removePilot = "RemovePilot",
            hoppedOutOf = "HoppedOutOf",
            setLocal = "SetLocal",
            isLocal = "IsLocal",
            isRemote = "IsRemote",
            setObjectiveOn = "SetObjectiveOn",
            setObjectiveOff = "SetObjectiveOff",
            getObjectiveName = "GetObjectiveName",
            setObjectiveName = "SetObjectiveName",
            setMaxHealth = "SetMaxHealth",
            setMaxAmmo = "SetMaxAmmo",
            setVelocity = "SetVelocity",
            setOmega = "SetOmega",
            setTransform = "SetTransform",
            setCritical = "SetCritical",
            setPosition = "SetPosition",
            setPilotClass = "SetPilotClass",
            setOwner = "SetOwner",
            setUserTarget = "SetUserTarget",
            setPerceivedTeam = "SetPerceivedTeam",
            setTeamNum = "SetTeamNum",
            isDamage = "IsDamage",
            isPerson = "IsPerson",
            isBuilding = "IsBuilding",
            isCraft = "IsCraft",
            isAliveAndPilot = "IsAliveAndPilot",
            isAlive = "IsAlive",
            isValid = "IsValid",
            removeObject = "RemoveObject",
            setLabel = "SetLabel", 
        },
        setHealth = function(self,fraction)
            self:setCurHealth(self:getMaxHealth()*fraction);
        end,
        setAmmo = function(self,fraction)
            self:setCurAmmo(self:getMaxAmmo()*fraction);
        end,
        getCommand = function(self)
            return AiCommand[self:getCurrentCommand()];
        end,
        genericFunc = function(self,func,...)
            return _G[func](self.h,...);
        end,
        genericGetFunc = function(self,var,func,...)
            if(IsValid(self.h)) then
                local v = self:genericFunc(func,...);
                rawset(self,var,v);
            end
            return rawget(self,var);
        end
    }
    
};
gameObject.mt = {__call = gameObject.new,__index = 
    function(t,v)
        local bzUtils = _G["bzUtils"];
        local gameObject = bzUtils.mHandle.gameObject;
        if(gameObject.prototype[v]) then
            return gameObject.prototype[v];
        elseif(t.genericGetMap[v]) then
            local x = t.genericGetMap[v];
            return function(self,...)
                return self:genericGetFunc(x[1],x[2],...);
            end
        elseif(t.genericMap[v]) then
            local x = t.genericMap[v];
            return function(self,...)
                return self:genericFunc(x,...);
            end
        end
    end
};
setmetatable(gameObject,gameObject.mt);

mHandle.Save = function(self)
    local mt = getmetatable(self);
    r = mt.__index.Save(self);
    r[0] = {};
    for i,v in pairs(self.objects) do
        table.insert(r[0],v:Save());
    end
    return r;
end
mHandle.Update = function(self,dtime)
    local mt = getmetatable(self);
    for i,v in pairs(self.objects) do
        v:Update(dtime);
    end
    mt.__index.Update(self,dtime);
end
mHandle.DeleteObject = function(self,handle)
    local mt = getmetatable(self);
    if(self.objects[handle]) then
        self.objects[handle]:DeleteObject();
        self.objects[handle] = nil;
    end
    mt.__index.DeleteObject(self);
end
mHandle.Load = function(self,data)
    local mt = getmetatable(self);
    local c = data[0];
    for i,v in pairs(c) do
        gameObject:Load(v);
    end
    mt.__index.Load(self,data);
end
mHandle.SaveDone = function(self)
    local mt = getmetatable(self);
    for i,v in pairs(self.objects) do
        v:SaveDone();
    end
    
    mt.__index.SaveDone(self);
end


mHandle.gameObject = gameObject;
bzUtils.mHandle = mHandle;

return mHandle;