--Storage module
--module for handling storage and table referencing,
--makes sure that reference to common tables are kept when saving and loading
--circular referencing is now supported!
--To use just require "mStore"

require "deepcopy";
--^To lacy to write my own: https://gist.githubusercontent.com/Deco/3985043/raw/ee4d5daf8d0f11e716164ed4bb2598e328a09c07/deepcopy.lua


local mStore = {
    currentSeq = 0
};

local refTable = {};
local seqTable = {};

local subH = {
    Save = _G["Save"],
    Load = _G["Load"]
};



local function getSeqnumber(t)
    local currentSeq = "@T_" .. tostring(mStore.currentSeq);
    if(not refTable[t]) then
        refTable[t] = {
            sqnum = currentSeq
        };
        seqTable[currentSeq] = t;
        mStore.currentSeq = mStore.currentSeq + 1;
    end
    return refTable[t];
end

local function Save()
    --Gets the storage table
    for k,v in pairs(mStore.seqTable) do
        mStore.seqTable[k] = nil;
    end
    for k,v in pairs(mStore.refTable) do
        mStore.refTable[k] = nil;
    end
    mStore.currentSeq = 0;
    print("Getting data!");
    --Doing a circular deepcopy to prevent current session from breaking!
    local allStore = table.deepcopy({subH["Save"]()},{"value_translate"},{
        userdata = function(stack, orig, copy, state, arg1, arg2)
            return orig, true;
        end
    });
    
    print("Generating ref table start")
    local childList = {allStore};
    while childList do
        local newChildList = nil;
        for i,v in pairs(childList) do
            for i2,v2 in pairs(v) do
                if(type(v2) == "table" and not v2.sqnum) then
                    if(not newChildList) then
                        newChildList = {};
                    end
                    v[i2] = getSeqnumber(v2);
                    table.insert(newChildList,v2);
               end
            end
        end
        childList = newChildList;
    end
    --Tree traversal, 
    --end nodes need to be traversed first!
    --Old, bad way. Did not support circular referencing 
    --[[
    local childList = {};
    for i,v in pairs(allStore) do
        table.insert(childList,{t=v,p=allStore,k=i});
    end
    local invertedTreeList = {};
    
    
    local maxDepth = 50;
    local cDepth = 0;
    while childList do
        table.insert(invertedTreeList,1,childList);
        local newChildList = nil;
        for i,v in pairs(childList) do
            if(type(v.t) == "table") then
                for i2,v2 in pairs(v.t) do
                    if(type(v2) == "table") then
                        local r = "-";
                        print(childList,r:rep(cDepth) .. "->",i2,v2);
                        if(not newChildList) then
                            newChildList = {};
                        end
                        table.insert(newChildList,{t=v2,p=v.t,k=i2});
                    end
                end
            end
        end
        childList = newChildList;
        cDepth = cDepth + 1;
        if(cDepth > maxDepth) then
            error("Tree was too deep!\nCircular referencing is not supported, make sure your data is clean!");
            break;
        end
    end
    for x,y in pairs(invertedTreeList) do
        for i,v in pairs(y) do
            print(y,i,v,getSeqnumber(v.t).sqnum);
            v.p[v.k] = getSeqnumber(v.t);
        end
    end
    --]]
    mStore.currentSeq = 0;
    return {
        refTable = allStore,
        defTable = mStore.seqTable
    };
end

local function Load(data)
    
    for k,v in pairs(mStore.seqTable) do
        mStore.seqTable[k] = nil;
    end
    for k,v in pairs(mStore.refTable) do
        mStore.refTable[k] = nil;
    end
    mStore.currentSeq = 0;


    local defTable = data.defTable;
    local refTable = data.refTable;
    
    local childList = refTable;
    for i,v in pairs(childList) do
        childList[i] = {t=v,p=refTable,k=i};
    end
    local cDepth = 0;
    while childList do
        local newChildList = nil;
        for i,v in pairs(childList) do
            local c = defTable[v.t.sqnum];
            v.p[v.k] = c;
            for i2,v2 in pairs(c) do
                local r = "-";
                if(type(v2) == "table" and v2.sqnum and type(v2.sqnum) == "string") then
                    if(not newChildList) then
                        newChildList = {};
                    end
                    table.insert(newChildList,{t=v2,k=i2,p=c});
                end
            end
        end
        childList = newChildList;
        cDepth = cDepth + 1;
    end    
    
    --print(unpack(refTable));
    subH["Load"](unpack(refTable));
end




--Hook to Load and Save
local hooks = {
    Save = Save,
    Load = Load
};

local p = {}; 
p.old = getmetatable(_G) or {};

p.__index = function(t,k)
    if(hooks[k]) then
        return hooks[k];
    elseif(p.old.__index) then
        return p.old.__index(t,k);
    else
        return rawget(t,k);
    end
end

p.__newindex = function(t,k,v)
    if(hooks[k]) then
        subH[k] = v;
    elseif(p.old.__newindex) then
        p.old.__newindex(t,k,v);
    else
        rawset(t,k,v);
    end
end

setmetatable(_G,p);

mStore.getSeqnumber = getSeqnumber;
mStore.Save = Save;
mStore.Load = Load;
mStore.hooks = hooks;
mStore.refTable = refTable;
mStore.seqTable = seqTable;



return mStore;