local bzUtils = require "mCore";
local mFlow = require "mFlow";
local mTimer = require "mTimer";


local mTask = bzUtils.core.mModule("mTask");
bzUtils:addModule(mTask);

local odfTaskTypes = {};
--Return values:
--0 nothing
--1 remove task, = task is done
--2 mission fail
--3 nothing critical, if all tasks are set
--to this and there is no timer, task is concidered success


odfTaskTypes["^kill%s+(.+)$"] = function(label)
    if(label==nil or type(label)~="string") then 
        error(string.format("kill task requires an object, was %s",tostring(label)));
    end
    return IsAlive(GetHandle(label)) and 0 or 1;
end

odfTaskTypes["^defend%s+(.+)$"] = function(label)
    if(label==nil or type(label)~="string") then 
        error(string.format("defend task requires an object, was %s",tostring(label)));
    end

    if(IsAlive(GetHandle(label))) then
        return 3;
    end
    return 2;
    --return IsAlive(GetHandle(label));
end

odfTaskTypes["get%s+(.+)%s+to%s+(.+)"] = function(label1,label2)
    if(((label1==nil or not type(label1)=="string")) or ((label2==nil or not type(label2)=="string")))  then 
        error(string.format("'get to' task requires two objects, was %s and %s",tostring(label1),tostring(label2)));
    end

    if(not (GetHandle(label1) and GetHandle(label2))) then 
        return 0;
    end
    return (GetDistance2D(GetHandle(label1),GetHandle(label2)) < 100) and 1 or 0;    
end

local odfTaskUpdate = mFlow.callback("odfTaskMasterUpdateFunction",function(task)
    --Check for tasks and remove them as they are finished
     if(task.subTasks) then
        local fail = false;
        local noMore = true;
        local allNonCritical = true;
        for i,v in pairs(task.subTasks) do
            local r = odfTaskTypes[v.func](unpack(v.args));
            noMore = false;
            if(r==1) then
                table.remove(task.subTasks,i);
            elseif(r == 2) then
                fail = true;
            elseif(r ~= 3) then
                allNonCritical = false;
            end    
        end
        if(fail) then
            task:fail();
        end
        if(noMore) then
            task:success();
        end
        if(allNonCritical and not task.countdownTimer) then
            task:success();
        end
    end
end);
local odfTaskStart = mFlow.callback("odfTaskMasterStartFunction",function(task)
    if(task.missionText) then
        AddObjective(task.missionFile,"white",8,task.missionText);
    end
    if(task.countdownTimer) then
        task.countdownTimer:start();
        if(task.showTimer) then
            task.countdownTimer:show();
        end
    end
end);

local odfTaskSuccess = mFlow.callback("odfTaskMasterSuccessFunction",function(task)
    if(task.missionText) then
        UpdateObjective(task.missionFile,task.sColor or "green",8,task.missionText);
    end
    if(task.success_task == "WIN") then
        SucceedMission(GetTime()+task.next_delay,task.win_text);
    else
        mTask:getTask(string.format("OTF_TASK_%s",task.success_task)):start();
    end
end);
local odfTaskFail = mFlow.callback("odfTaskMasterFailFunction",function(task)
    if(task.missionText) then
        UpdateObjective(task.missionFile,task.fColor or "red",8,task.missionText);
    end
    if(task.fail_task == "FAIL") then
        print(task.next_delay,task.fail_text);
        FailMission(GetTime()+task.next_delay,task.fail_text);
    else
        mTask:getTask(string.format("OTF_TASK_%s",task.fail_task)):start();
    end
end);

local odfTaskFinish = mFlow.callback("odfTaskMasterFinishFunction",function(task)
    if(task.countdownTimer) then
        task.countdownTimer:stop();
        task.countdownTimer:kill();
        task.countdownTimer = nil;
    end
    --Clean up the rest!
end);
local odfTaskStop = mFlow.callback("odfTaskMasterStopFunction",function(task)
    if(task.countdownTimer) then
        task.countdownTimer:stop();
    end
end);
local odfTaskResume = mFlow.callback("odfTaskMasterResumeFunction",function(task)
    if(task.countdownTimer) then
        task.countdownTimer:start();
    end
    --Clean up the rest!
end);
local odfTaskCountdown = mFlow.callback("odfTaskMasterCountdownTimeout",function(timer)
    if(timer.task) then
        if(timer.condition == "SUCCESS") then
            timer.task:success();
        else
            timer.task:fail();
        end
    end
end);


local odfTaskMasterEvents = {
    finish = "odfTaskMasterFinishFunction",
    start = "odfTaskMasterStartFunction",
    stop = "odfTaskMasterStopFunction",
    resume = "odfTaskMasterStopFunction",
    success = "odfTaskMasterSuccessFunction",
    fail = "odfTaskMasterFailFunction",
    update = "odfTaskMasterUpdateFunction"
            
};
local task = {
    new = function(cls,name,def)
        if(mTask.tasks[name]) then
            return mTask.tasks[name];
        end
        local user_store = {};
        if(def) then
            user_store = def.user_store;
        end
        user_store = setmetatable(user_store,cls:mt());
        local inst = def or {};
        if(not def) then
            inst.name = name;
            inst.dead = false;
            inst.running = false;
            inst.init = false;
            inst.eventHandler = mFlow.eventHandler(inst);
        end
        mTask.tasks[name] = inst;
        return setmetatable(inst,{__index = user_store,__newindex = user_store});
    end,
    Load = function(cls,data)
        
        local inst = cls:new(data.name,data);
        mFlow.eventHandler:Load(inst.eventHandler,inst);
        local uStore = inst["user_store"];
        rawset(inst,"user_store",nil);
        for i,v in pairs(uStore) do
            inst[i] = v;
        end
        return inst;
    end,
    prototype = {
        saveable = true,
        addListener = function(self,mEventName,mCallback,...)
            self.eventHandler:addListener(mEventName,mCallback,...);
        end,
        Save = function(self)
            local mt = getmetatable(self);
            if(not rawget(self,"_currentSave")) then
                rawset(self,"_currentSave",{
                    name = self.name,
                    eventHandler = self.eventHandler:Save(),
                    dead = self.dead,
                    running = self.running,
                    init = self.init,
                    user_store = {}
                });
                for i,v in pairs(mt.__index) do
                    local item = v;
                    if(type(v) == "table" and v.saveable and v.Save) then
                        item = v:Save();
                    end
                    rawget(self,"_currentSave")["user_store"][i] = item;
                end
            end
            return rawget(self,"_currentSave");
        end,
        isCurrent = function(self)
            return self.running;
        end,
        isDead = function(self)
            return self.dead;
        end,
        kill = function(self)
            self.dead = true;
        end,
        SaveDone = function(self)
            rawset(self,"_currentSave",nil);
        end,
        Update = function(self,dtime)
            self.eventHandler:dispatchEvent(mFlow.event("update",dtime));
        end,
        AddObject = function(self,handle)
            self.eventHandler:dispatchEvent(mFlow.event("addObject",handle));    
        end,
        start = function(self)
            self.running = true;
            if(not self.init) then
                self.eventHandler:dispatchEvent(mFlow.event("start"));
                self.init = true;
            else
                self.eventHandler:dispatchEvent(mFlow.event("resume"));
            end
        end,
        stop = function(self)
            self.running = false;
            self.eventHandler:dispatchEvent(mFlow.event("stop"));
        end,
        finish = function(self)
            if(not self:isDead()) then
                self.running = false;
                self.eventHandler:dispatchEvent(mFlow.event("finish"));
                self:kill();
            end
        end,
        success = function(self)
            if(not self:isDead()) then
                self.eventHandler:dispatchEvent(mFlow.event("success"));
                self:finish();
            end
        end,
        fail = function(self)
            if(not self:isDead()) then
                self.eventHandler:dispatchEvent(mFlow.event("fail"));
                self:finish();
            end
        end
    },
    mt = function(cls)
        return {__index = cls.prototype,__call = cls.new};
    end
    
};
setmetatable(task,task:mt());



mTask.task = task;
mTask.AddObject = function(self,handle)
    local mt = getmetatable(self);
    for i,v in pairs(self.tasks) do
        if(v:isCurrent()) then
            v:AddObject(handle);
        end
    end
    mt.__index.Update(self,dTime);
end
mTask.Update = function(self,dTime)
    local mt = getmetatable(self);
    for i,v in pairs(self.tasks) do
        if(v:isCurrent()) then
            v:Update(dTime);
        end
        if(v:isDead()) then
            self.tasks[i] = nil;
        end
    end
    mt.__index.Update(self,dTime);
end


mTask.Save = function(self)
    local mt = getmetatable(self);
    r = mt.__index.Save(self);
    r[0] = {};
    for i,v in pairs(self.tasks) do
        table.insert(r[0],v:Save());
    end
    return r;
end

mTask.Load = function(self,data)
    local mt = getmetatable(self);
    local c = data[0];
    for i,v in pairs(c) do
        task:Load(v);
    end
    mt.__index.Load(self,data);
end

mTask.SaveDone = function(self)
    local mt = getmetatable(self);
    for i,v in pairs(self.tasks) do
        v:SaveDone();
    end
    
    mt.__index.SaveDone(self);
end

mTask.getTask = function(self,taskName)
    if(not self.tasks[taskName]) then
        error(string.format("No task with the name %s exists!",tostring(taskName)),3);
    end
    return self.tasks[taskName];
end

local function generateFromJSON(jsonfile)
    
    
end

local function generateFromOdf(odf)
    --Generates mission from an odf
    print("Generating tasks!",odf);
    local odfFile = bzUtils.core.odfFile(odf);
    local mainHeader = odfFile["Main"];
    --print(odfFile.getHeader,odfFile.getHeader("Main"));
    local objective_count = mainHeader:getAsInt("objective_count");
    if(objective_count == nil) then
        error(string.format("\"objective_count\" was not found in \"Main\"!"));
    end
    local objectives = {};
    for i=1,objective_count do
        --Generate each objective
        local objName = mainHeader[string.format("objective_%d",i)];
        if(not objName) then
            error(string.format("\"objective_%d\" was not found in \"Main\"!",i));
        end
        local task = mTask.task(string.format("OTF_TASK_%s",objName));
        local tSection = odfFile[objName];
        local task_count = tSection:getAsInt("task_count");
        task.subTasks = {};
        if(task_count <= 0) then
            task.subTasks = nil;
        end
        for i=1, task_count do
            local tText = tSection[string.format("task_%d",i)];
            local fName = nil;
            local mtch = nil;
            for x,y in pairs(odfTaskTypes) do
                mtch = {tText:find(x)};
                if(mtch[1]) then
                    table.remove(mtch,2);
                    table.remove(mtch,1);
                    fName = x;
                    break;
                end
            end
            if(not fName) then
                error(string.format("%s is not a valid task!",tText));
            end
            table.insert(task.subTasks,{
                t = tText,
                args = mtch,
                func = fName
            });
        end
        for i,v in pairs(odfTaskMasterEvents) do
            task:addListener(i,v);
        end
        task.missionFile = tSection["task_name"] or tSection["task_otf"] or objName;
        task.missionText = tSection["task_text"] or UseItem(task.missionFile);
        if(task.missionText) then
            task.missionText = task.missionText:gsub("\\n","\n");
        end
        local customEvents = {
            finish = tSection["on_finish"],
            start = tSection["on_start"],
            stop = tSection["on_stop"],
            success = tSection["on_success"],
            fail = tSection["on_fail"],
            update = tSection["on_update"],
            resume = tSection["on_resume"],
            addObject = tSection["on_addObject"]
        };
        for i,v in pairs(customEvents) do
            task:addListener(i,v);
        end
        task.sColor = tSection["success_color"];
        task.fColor = tSection["fail_color"];
        
        local timelimit = tSection:getAsInt("timelimit");
        if(timelimit ~= 0) then
            task.countdownTimer = mTimer.timer(timelimit,false);
            task.countdownTimer.task = task;
            task.countdownTimer.condition = tSection["time_condition"];
            task.countdownTimer:addListener("timeout","odfTaskMasterCountdownTimeout");
            task.showTimer = tSection:getAsBool("timeshow");
        end
        task.next_delay = tSection:getAsInt("end_delay");
        task.win_text = tSection["win_text"];
        task.fail_text = tSection["fail_text"];
        
        task.success_task = tSection["objective_success"];
        task.fail_task = tSection["objective_fail"];
        table.insert(objectives,task);
    end
    for i,v in pairs(objectives) do
        local f_test = (v.fail_task == "FAIL") or mTask:getTask(string.format("OTF_TASK_%s",v.fail_task));
        local s_test = (v.success_task == "WIN") or mTask:getTask(string.format("OTF_TASK_%s",v.success_task));
    end
    local first_objective = mainHeader["first_objective"];
    if(not first_objective) then
        error("\"first_objective\" was not found in \"Main\"!");
    end
    return {
        start = function(t)
            print("Starting first objective");
            t.first_objective:start();
        end,
        first_objective = mTask:getTask(string.format("OTF_TASK_%s",first_objective)),
        objectives = objectives
    };
end

mTask.generateFromOdf = generateFromOdf;
mTask.tasks = {};


bzUtils.mTask = mTask;
return mTask;

