bzUtils = require "mCore";
mFlow = require "mFlow";

local mTimer = bzUtils.core.mModule:new("mTimer");
bzUtils:addModule(mTimer);

local timer = {
    --[[
        keepTrackOf defaults to true and makes sure the timer
        is saved, only set to false, when intending to control
        saving/loading of the timer yourself
    ]]
    new = function(cls,t,loop,keepTrackOf,def)
        local user_store = setmetatable({},cls:mt());        
        local tm = def or {};
        tm.track = --[[keepTrackOf or]] true;
        tm.p = t;
        tm.eventHandler = mFlow.eventHandler(tm);
        tm.c_time = 0;
        tm.loop = loop or false;
        tm.run = false;
        tm.dead = false;
        tm.showTimer = false;
        setmetatable(tm,{__index = user_store,__newindex = user_store});
        table.insert(mTimer.timers,tm);
        return tm;
    end,
    Load = function(cls,data)
        --Makes the timer auto load for everyone and keeps reference
        print(data);
        local sCopy = {};
        for i,v in pairs(data) do
            sCopy[i] = v;
            data[i] = nil;
        end
        
        inst = cls:new(0,false,true,data);
        
        rawset(inst,"run",sCopy["running"]);
        rawset(inst,"p",sCopy["period"]);
        rawset(inst,"c_time",sCopy["accTime"]);
        rawset(inst,"dead",sCopy["dead"]);
        rawset(inst,"loop",sCopy["loop"]);
        rawset(inst,"eventHandler",mFlow.eventHandler:Load(sCopy["eHandler"],inst));
 
        --self.eventHandler:Load(data["eHandler"]);
        for i,v in pairs(sCopy["user_store"]) do
            inst[i] = v;
        end
        if(sCopy["display"]) then
            inst:show();
        end
        return inst;
        --self.eventHandler:dispatchEvent(mFlow.event("load",data["user_store"]));
    end,
    prototype = {
        saveable = true,
        addListener = function(self,mEventName,mCallback)
            self.eventHandler:addListener(mEventName,mCallback);
        end,
        Save = function(self)
            local mt = getmetatable(self);
            --Make sure that reference is kept(aka prevents duping)
            if not rawget(self,"_currentSave") then
                rawset(self,"_currentSave",{
                    running = self.run,
                    period = self.p,
                    accTime = self.c_time,
                    dead = self.dead,
                    loop = self.loop,
                    display = self.showTimer,
                    eHandler = self.eventHandler:Save(),
                    user_store = {}
                });
                for i,v in pairs(mt.__index) do
                    local item = v;
                    if(type(v) == "table" and v.saveable and v.Save) then
                        item = v:Save();
                    end
                    rawget(self,"_currentSave")["user_store"][i] = item;
                end
            end
            return rawget(self,"_currentSave");
            
        end,
        SaveDone = function(self)
            rawset(self,"_currentSave",nil);
            self.eventHandler:SaveDone();
        end,
        kill = function(self)
            if(self.showTimer) then
                self:hide();
            end
            self.dead = true;
            self.eventHandler:dispatchEvent(mFlow.event("destroy"));
        end,
        isDead = function(self)
            return self.dead;
        end,
        start = function(self)
            self.run = true;
            self.eventHandler:dispatchEvent(mFlow.event("start"));
        end,
        stop = function(self)
            self.run = false;
            self.eventHandler:dispatchEvent(mFlow.event("stop"));    
        end,
        show = function(self)
            if(mTimer.dispTimer) then
                mTimer.dispTimer:hide();
            end
            mTimer.dispTimer = self;
            self.showTimer = true;
        end,
        hide = function(self)
            if(self.showTimer) then
                mTimer.dispTimer = nil;
                StartCockpitTimer(0);
                StopCockpitTimer();
                HideCockpitTimer();
                self.showTimer = false;
            end
        end,
        tick = function(self,dTime)
           if(self.run) then
                self.c_time = self.c_time + dTime;
                if(self.c_time >= self.p) then
                    if(not self.loop) then
                        print("Killing timer",self);
                        self:kill();
                    end
                    self.run = self.loop;
                    self.c_time = 0;
                    self.eventHandler:dispatchEvent(mFlow.event("timeout"));
                end
            end
            if(self.showTimer) then
                StartCockpitTimer( math.floor((self.p+0.99) - self.c_time));
            end
        end
    },
    mt = function(cls)
        return {__index = cls.prototype,__call = cls.new}; 
    end
};
setmetatable(timer,timer:mt());
mTimer.timer = timer;
mTimer.timers = {};
mTimer.dispTimer = nil;

mTimer.Update = function(self,dTime)
    local mt = getmetatable(self);
    for i,v in pairs(self.timers) do
        v:tick(dTime);
        if(v:isDead()) then
            table.remove(self.timers,i);
        end
    end
    mt.__index.Update(self,dTime);
end

mTimer.Save = function(self)
    local mt = getmetatable(self);
    r = mt.__index.Save(self);
    r[0] = {};
    for i,v in pairs(self.timers) do
        if(v.track) then
            table.insert(r[0],v:Save());
        end
    end
    return r;
end
mTimer.Load = function(self,data)
    local mt = getmetatable(self);
    local c = data[0];
    for i,v in pairs(c) do
        --local t = timer(0,false,true);
        timer:Load(v);
    end
    mt.__index.Load(self,data);
end
mTimer.SaveDone = function(self)
    local mt = getmetatable(self);
    for i,v in pairs(self.timers) do
        v:SaveDone();
    end
    
    mt.__index.SaveDone(self);
end

bzUtils.mTimer = mTimer;

return mTimer;