--[[
    Module for controlling unit behaviour
--]]

local bzUtils = require "mCore";
local mHandle = require "mHandle";
local mFlow = require "mFlow"

local mUnit = bzUtils.core.mModule("mUnit");
mHandle:addModule(mUnit);

local controller = {
    new = function(cls,def)
        local inst = setmetatable(def or {},cls:mt());
        if(not def) then
            inst.eventHandler = mFlow.eventHandler(inst);
            inst.unitListeners = {};
            inst.a_odfs = {};
            inst.a_classes = {};
            inst.units = {};
        end
        table.insert(mUnit.controllers,inst);
        return bzUtils.core.createLocalStore(inst);
    end,
    Load = function(cls,data)
        local cData = {};
        for i,v in pairs(data) do
            cData[i] = v;
            data[i] = nil;
        end
        data.eventHandler = mFlow.eventHandler:Load(cData.eHandler,data);
        data.unitListeners = cData.uList;
        data.a_odfs = cData.odfs;
        data.a_classes = cData.classes;
        data.units = cData.units;
        local inst = cls:new(data);
        local uStore = cData["user_store"];
        for i,v in pairs(uStore) do
            inst[i] = v;
        end
        return inst;
    end,
    prototype = {
        saveable = true,
        Save = function(self)
            if(not rawget(self,"_currentSave")) then
                local r = {
                    eHandler = self.eventHandler:Save(),
                    uList = self.unitListeners,
                    odfs = self.a_odfs,
                    classes = self.a_classes,
                    units = self.units;
                };
                rawset(self,"_currentSave",r);
                bzUtils.core.createSaveTable(self);
            end
            return rawget(self,"_currentSave");
        end,
        SaveDone = function(self)
            rawset(self,"_currentSave",nil);
        end,
        addListener = function(self,mEventName,mCallback,...)
            self.eventHandler:addListener(mEventName,mCallback,...);
        end,
        dispatchEvent = function(self,mEvent)
            self.eventHandler:dispatchEvent(mEvent);
        end,
        addUnitListener = function(self,mEventName,mCallback)
            local c_name = ""
            if(type(mCallback) == "table") then
                c_name = mCallback.name;
            else
                if(not mFlow:getCallback(mCallback)) then
                    error("No callback named " .. tostring(mCallback) .. " exists!",3);
                end
                c_name = mCallback;
            end
            table.insert(self.unitListeners,{mEventName,mCallback});            
        end,
        applyToODF = function(self,odf,anyNation)
            self.a_odfs[odf:gmatch("[^%c]+")()] = {anyNation};
        end,
        applyToClass = function(self,class)
            self.a_classes[class] = true;
        end,
        runOnAll = function(self)
            for v in AllObjects() do
                self:AddObject(v);
            end
        end,
        getUnits = function(self)
            local ret = {};
            for i,v in pairs(self.units) do
                if(not v.dead) then
                    table.insert(ret,v);
                end
            end
            self.units = ret;
            return ret;
        end,
        AddObject = function(self,handle)
            if((not IsNetGame()) or (not IsRemote(handle))) then
                local testRes = self.eventHandler:dispatchEvent(mFlow.event("handle_test",handle));
                local odf = GetOdf(handle):gmatch("[^%c]+")();
                local any = self.a_odfs[odf:sub(2,odf:len())];
                if(self.a_odfs[odf] or (any and any[1]) or self.a_classes[GetClassLabel(handle)]) then
                    table.insert(testRes,{true});
                end 
                for i,v in pairs(testRes) do
                    if(v[1]) then
                        local obj = mHandle.gameObject(handle);
                        for k,v2 in pairs(self.unitListeners) do
                            obj:addListener(v2[1],v2[2]);
                        end
                        table.insert(self.units,obj);
                        self.eventHandler:dispatchEvent(mFlow.event("addObject",obj));
                        break;
                    end
                end
            end
        end
    };
    mt = function(cls)
       return {__call = cls.new, __index = cls.prototype};
    end
    
};
setmetatable(controller,controller:mt());
mUnit.controller = controller;
mUnit.controllers = {};


mUnit.AddObject = function(self,handle)
    local mt = getmetatable(self);
    for i,v in pairs(self.controllers) do
        v:AddObject(handle);
    end
    mt.__index.AddObject(self,handle);
end

mUnit.Save = function(self)
    local mt = getmetatable(self);
    r = mt.__index.Save(self);
    r[0] = {};
    for i,v in pairs(self.controllers) do
        table.insert(r[0],v:Save());
    end
    return r;
end

mUnit.Load = function(self,data)
    local mt = getmetatable(self);
    local c = data[0];
    for i,v in pairs(c) do
        controller:Load(v);
    end
    mt.__index.Load(self,data);
end
mUnit.SaveDone = function(self)
    local mt = getmetatable(self);
    for i,v in pairs(self.controllers) do
        v:SaveDone();
    end
    mt.__index.SaveDone(self);
end

mHandle.mUnit = mHandle;

return mUnit;