local bzUtils = require "mCore";
local mFlow = require "mFlow";
local mTimer  = require "mTimer";
local mHandle = require "mHandle";
local mUnit = require "mUnit";
local mTask = require "mTask";

SetAIControl(2, true);

function init()
    mFlow.callback("dropship_add",function(c,ship)
        ship.cargo = {};
        ship.com_queue = {};
        ship.cam = false;
        ship.dep = true;
        local t = GetTerrainHeightAndNormal(ship:getPosition());
        local h = ship:getPosition().y - t;
        ship.onGround = false; 
        print("add_dship:",h);
    end);
    mFlow.callback("dropship_update",function(ship,dtime)
        local t = GetTerrainHeightAndNormal(ship:getPosition());
        local h = ship:getPosition().y - t;
        ship.onGround = (not ship.dep) and (not ship:isDeployed()) and h < 10; 
        if(h > 30) then
            ship.dep = false;
        end
        if(ship.cam) then
            local rVec = DotProduct(ship:getFront(),Normalize(ship:getVelocity())) * Length(ship:getVelocity());
            CameraObject(ship.h,0,4000 - ship:getVelocity().y*10,-8000-rVec*20,ship.h);
        end
        local c = 0;
        for i,v in pairs(ship.cargo) do
            SetTransform(i,ship:getTransform());
            SetPosition(i,ship:getPosition() + SetVector(0,2000 + 50*c,0) );
            SetVelocity(i,ship:getVelocity());
            SetOmega(i,ship:getOmega());
            c = c+1;
        end
        if((ship:getCurrentCommand() == AiCommand["NONE"]) and #ship.com_queue >= 1) then
            ship:setCommand(unpack(ship.com_queue[1]));
            table.remove(ship.com_queue,1);
        end
    end);   
    mFlow.callback("dropship_command",function(ship,new,old)
        print("Command:",AiCommand[new],AiCommand[old],ship.onGround);
        
        if( (new == AiCommand["GO"]) and ship.onGround) then
            ship:deploy();
        end
        if((new == AiCommand["DROPOFF"]) and not(ship.onGround)) then
            ship:deploy();
            ship:stop(0);
        end
    end);
    mFlow.callback("dropship_who",function(ship,new,old)
        if(ship:getCurrentCommand() == AiCommand["GO"] and ship.onGround) then
            ship:deploy();
        end
    end);
    
    mFlow.callback("dropship_receive",function(ship,from,what,...)
        print(receive,from,what,...);
        local who = ...;
        if(what == "PICKUP" and who) then
            if(GetPlayerHandle() == who) then
                ship.cam = CameraReady();
            end
            ship.cargo[who] = true;
        end
        if(what == "DROPOFF") then
            if(ship.cam) then
                CameraFinish();
                ship.cam = false;
            end
            ship.cargo = {};
        end
        if(what == "QUEUE") then
            table.insert(ship.com_queue,{...});
        end
    end);
    mFlow.callback("intro_start",function(task)
        SetObjectiveOn(GetHandle("relic_hangar"));
        task.cCam = 0;
        task.timeAcc = 0;
        task.cObj = 0;
        task.tfile = UseItem("text1.otf");
        CameraReady();
    end);
    mFlow.callback("intro_update",function(task,dtime)
        if(CameraCancelled()) then
            CameraFinish();
            task.cCam = 2;
        elseif(task.cCam == 0) then
            if(CameraPath("camPath0",3000,1000,GetHandle("svrecy1"))) then
                self.cCam = 1;
            end
        elseif(task.cCam == 1) then
            if(CameraPath("camPath1",3000,700,GetHandle("relic_hangar"))) then
                self.cCam = 2;
                CameraFinish();
            end
        elseif(task.cCam == 2) then
            task.timeAcc = task.timeAcc + dtime;
            if(task.cObj == 0 and task.timeAcc >= 2) then
                AddObjective("chan_open","blue",8,"Opening communication channel");
                task.timeAcc = 0;
                task.cObj = task.cObj + 1;
            elseif(task.cObj == 1) then
                UpdateObjective("chan_open","blue",8,"Opening communication channel" .. ("."):rep((task.timeAcc*2) % 3));
                if(task.timeAcc > 4) then
                    UpdateObjective("chan_open","green",8,"Connection established!\nTransmission started");
                    AddObjective("t_text","blue",8,"");
                    task.timeAcc = 0;
                    task.cObj = task.cObj + 1;
                end
            elseif(task.cObj == 2) then
                local eIndex = math.floor(task.timeAcc*20);
                UpdateObjective("t_text","blue",10,task.tfile:sub(1,eIndex));
                if(eIndex >= task.tfile:len()) then
                    RemoveObjective("chan_open");
                    task.timeAcc = 0;
                    task.cObj = 3;
                end
            elseif(task.cObj == 3 and task.timeAcc > 10) then
                RemoveObjective("t_text");
                AddObjective("obj1.otf","white",8);
                AddObjective("obj1d.otf","white",8);
                task:finish();
                task.timeAcc = 0;
                task.cObj = 4; 
            end
        end
    end);
    mFlow.callback("obj_attack_start",function(task)
        task.recy = GetHandle("svrecy1");
        task.muf = GetHandle("svmuf1")
    end);
    mFlow.callback("obj_attack_update",function(task,dtime)
        if((not IsAlive(task.recy)) and (not IsAlive(task.muf))) then
            task:success();
        end
    end);
    mFlow.callback("obj_attack_success",function(task)
        UpdateObjective("obj1.otf","green",8);
        local t1 = mTimer.timer(10,false);
        t1:addListener("timeout","obj_twait_timeout");
        t1.task = task.n_task;
        t1:start();
        local t2 = mTimer.timer(180,false);
        t2:addListener("timeout","spawn_wait_timeout1");
        t2.other = task.n_task;
        t2:start();
        AddObjective("obj2.otf","white",8);
    end);
    mFlow.callback("obj_twait_timeout",function(timer)
        timer.task:start();
    end);
    mFlow.callback("spawn_wait_timeout1",function(timer)
        local t2 = mTimer.timer(2,true);
        t2:addListener("timeout","vehicle_spawn_timeout");
        t2.bList = {
			{"svrecy",1},{"svmuf",1},{"svscav",3},{"svwalk",5},{"svtank",15},{"svfigh",5}
		};
        t2.i = 1;
        t2.itemL = t2.bList[1][2];
        t2:start();
        timer.other.cCam = 1;
    end);
    mFlow.callback("vehicle_spawn_timeout",function(timer)
        local cObj = timer.bList[timer.i][1];
        local h = BuildObject(cObj,2,BuildPositionRotationMatrix(0,3.14/2,0,GetPosition(timer.sp or "lPad_spawn")));	
        if(GetClassLabel(h) == "wingman") then
            Defend2(h,timer.tar or GetHandle("relic_hangar"));
        end
        timer.itemL = timer.itemL - 1;
        if(timer.itemL <= 0 ) then
            timer.i = timer.i + 1;
            if(timer.bList[timer.i] ~= nil) then
                timer.itemL = timer.bList[timer.i][2];
            else
                timer:kill();
            end
        end
    end);
    mFlow.callback("obj_def_fail",function(task)
        UpdateObjective("obj1d.otf","red",8);
        FailMission(GetTime()+5);
    end);
    mFlow.callback("obj_def_success",function(task)
        UpdateObjective("obj1d.otf","green",8);
    end);
    
    
    mFlow.callback("obj_deconstruct_start",function(task)
        task.cCam = 0;
        local p = GetPosition("a_spawn");
        local recy = BuildObject("avrecy",1,GetPositionNear(p,50,100));
		local h1 = BuildObject("avtanf",1,GetPositionNear(p,50,100));
		local h2 = BuildObject("avtanf",1,GetPositionNear(p,50,100));
		local h3 = BuildObject("avtanf",1,GetPositionNear(p,50,100));
		SetScrap(1,100);
		Goto(recy,GetHandle("recy_g2"),0);
        Follow(h1,recy,0);
        Follow(h2,recy,0);
        Follow(h3,recy,0);
	
    end);
    mFlow.callback("obj_deconstruct_update",function(task,dtime)
        if(task.cCam == 1) then
            CameraReady();
            task.cCam = 2;
        end
        if(task.cCam == 2) then
            if(CameraPath("camPath2",300,700,GetHandle("lPad")) or CameraCancelled()) then
                self.cCam = 3;
                CameraFinish();
            end
        end
    end);
    mFlow.callback("obj_deconstruct_success",function(task)
        UpdateObjective("obj2.otf","green",8);
        SetObjectiveOff(GetHandle("relic_hangar"));
        SetObjectiveOn(task.relic);
        local t2 = mTimer.timer(2,true);
        t2:addListener("timeout","vehicle_spawn_timeout");
        t2.bList = {
		    {"svhraz",10},{"svtank",50},{"svfigh",20}
        };
        t2.sp = "sp1";
        t2.i = 1;
        t2.itemL = t2.bList[1][2];
        t2.tar = task.relic;
        t2:start();
        
        local t3 = mTimer.timer(2,true);
        t3:addListener("timeout","vehicle_spawn_timeout");
        t3.bList = {
		    {"svhraz",2},{"svtank",3},{"svfigh",5}
        };
        t3.sp = "a_spawn";
        t3.i = 1;
        t3.itemL = t2.bList[1][2];
        t3.tar = task.relic;
        t3:start();
        
        
        
        task.n_task.relic = task.relic;
        task.n_task:start();
    end);
    mFlow.callback("constructor_add",function(c,cnst)
        --cnst.ctar = false;
        cnst.recyTarget = nil;
        cnst.recyAsk = false;
    end);
    mFlow.callback("constructor_update",function(cnst,dtime)
        if(cnst:isDeployed() and cnst.recyTarget and not(cnst.recyAsk)) then
            cnst:send(cnst.recyTarget,"SP_RECY");
            cnst.recyAsk = true;
        end
    end);
    mFlow.callback("constructor_command",function(cnst,new,old)
        if((new == AiCommand["PICKUP"]) and cnst.recyTarget) then
            cnst:send(cnst.recyTarget,"RECY_STOP");
            cnst.recyTarget = nil;
            cnst.recyAsk = false;
        end
    end);
    mFlow.callback("constructor_who",function(cnst,new,old)
        --print("Who:",GetLabel(new),GetLabel(old));
        if((cnst:getCurrentCommand() == AiCommand["RECYCLE"]) and new) then
            cnst.recyTarget = mHandle.gameObject(new);
            cnst.recyAsk = false;
        end
    end);
    mFlow.callback("constructor_receive",function(cnst,from,what,...)
        if(from == cnst.recyTarget) then
            if(what == "DONE_RECY") then
                cnst:pickup();
            elseif(what == "SP_RECY") then
                cnst:stop();
            end
        end
    end);
    mFlow.callback("constructor_destroy",function(cnst)
        if(cnst.recyTarget) then
            cnst:send(cnst.recyTarget,"RECY_STOP");
        end
    end);
    mFlow.callback("obj_htimer_timeout",function(timer)
        --spawn relic, etc etc
        local hangar = timer.h;
        
        local p = hangar:getPosition();
        p.y = 100000;
        p.y = GetFloorHeightAndNormal(p);
        hangar.stask.relic = BuildObject("obdata",0,p);
        
        hangar.done = true;
        hangar:send(hangar.cnst,"DONE_RECY");
        hangar.ftask:success();
        hangar.stask:success();
    end);
    mFlow.callback("obj_hangar_receive",function(hangar,from,what)
        if(from) then
            if((what == "SP_RECY") and (not hangar.cnst) and (not hangar.done)) then
                hangar:send(from,"SP_RECY");
                hangar.hTimer:start();
                hangar.hTimer:show();
                hangar.cnst = from;
            end
            if(from == hangar.cnst) then
                if(what == "RECY_STOP") then
                    hangar.hTimer:stop();
                    hangar.hTimer:hide();
                    hangar.cnst = nil;
                end
            end
        end
    end);
    mFlow.callback("obj_hangar_destroy",function(hangar)
        if(hangar.cnst) then
            hangar:send(hangar.cnst,"DONE_RECY");
        end
        if(not hangar.done) then
            hangar.ftask:fail();
        end
    end);
    mFlow.callback("obj_escort_start",function(task)
        task.dship = nil;
        task.rpick = false;
        task.ppick = false;
        local nav = BuildObject("apcamr",1,"dustoff");
        SetObjectiveName(nav,"Dustoff site");
        task.tText = UseItem("obj3.otf");
        task.timeAcc = 0;
        AddObjective("escort","blue",8,"");
        
    end);
    mFlow.callback("obj_escort_update",function(task,dtime)
        local i = 1 + math.floor(task.timeAcc*20);
        if(i < task.tText:len()) then
            UpdateObjective("escort","blue",8,task.tText:sub(1,i));
        end
        task.timeAcc = task.timeAcc + dtime;
        if((not task.dship) and (GetDistance(task.relic,"dustoff") < 300 or GetDistance(GetPlayerHandle(),"dustoff") < 300) ) then
            task.dship = mHandle.gameObject(BuildObject("avdrop",1,"dship_spawn"));
            --task.dship:goto("dustoff");
            task.dship:message(nil,"QUEUE",AiCommand["GO"],0,nil,"dship_ppath");
            task.dship:message(nil,"QUEUE",AiCommand["DROPOFF"],0);
            --Goto(task.dship,"dustoff");
        end
        if(task.dship and task.dship.onGround) then
            if(not task.rpick) then
                if(task.dship:getDistance(task.relic) < 100) then
                    print("PICKUP relic");
                    task.dship:message(nil,"PICKUP",task.relic);
                    if(GetTug(task.relic)) then
                        task.dship:message(nil,"PICKUP",GetTug(task.relic));
                    end
                    task.rpick = true;
                end
            elseif(not task.ppick) then
                if(task.dship:getDistance(GetPlayerHandle()) < 100) then
                    print("PICKUP player");
                    task.dship:message(nil,"PICKUP",GetPlayerHandle());
                    task.ppick = true;
                end
            end
        end
        if(task.ppick and task.rpick) then
            task:success();
        elseif(not IsValid(task.relic)) then
            task:fail();
        end
    end);
    mFlow.callback("obj_escort_success",function(task)
        task.dship:goto("dropship_path",0);
        UpdateObjective("escort","green",8,task.tText);
        SucceedMission(GetTime()+20);
    end);
    mFlow.callback("obj_escort_fail",function(task)
        UpdateObjective("escort","red",8,task.tText);
        FailMission(GetTime()+5);
    end);
    mFlow.callback("mem_timeout",function(timer)
        print(("Mem usage: %sKB"):format(collectgarbage("count")));
    end);
end


function Start()

    SetMaxScrap(2,1000);
    SetMaxScrap(1,45);
    
    local intro_task = mTask.task("intro_task");
    intro_task:addListener("start","intro_start");
    intro_task:addListener("update","intro_update");
    
    local attack_task = mTask.task("attack_task");
    attack_task:addListener("start","obj_attack_start");
    attack_task:addListener("update","obj_attack_update");
    attack_task:addListener("success","obj_attack_success");
    
    local def_task = mTask.task("def_task");
    def_task:addListener("fail","obj_def_fail");
    def_task:addListener("success","obj_def_success");
    
    
    local recy_task = mTask.task("recy_task");
    recy_task:addListener("start","obj_deconstruct_start");
    recy_task:addListener("update","obj_deconstruct_update");
    recy_task:addListener("success","obj_deconstruct_success");
    
    local escort_task = mTask.task("escort_task");
    escort_task:addListener("start","obj_escort_start");
    escort_task:addListener("update","obj_escort_update");
    escort_task:addListener("success","obj_escort_success");
    escort_task:addListener("fail","obj_escort_fail");
    
    
    attack_task.n_task = recy_task;
    recy_task.n_task = escort_task;
    
    
    local cnst_controller = mUnit.controller();
    cnst_controller:addListener("addObject","constructor_add");
    cnst_controller:addUnitListener("update","constructor_update");
    cnst_controller:addUnitListener("command_changed","constructor_command");
    cnst_controller:addUnitListener("who_changed","constructor_who");
    cnst_controller:addUnitListener("receive","constructor_receive");
    cnst_controller:addUnitListener("delete","constructor_destroy");
    cnst_controller:applyToODF("avcnst");
    cnst_controller:runOnAll();
    
    local hangar = mHandle.gameObject(GetHandle("relic_hangar"));
    hangar:addListener("delete","obj_hangar_destroy");
    hangar:addListener("receive","obj_hangar_receive");
    
    hangar.hTimer = mTimer.timer(60*5,false);
    hangar.hTimer:addListener("timeout","obj_htimer_timeout");
    hangar.hTimer.h = hangar;
    hangar.done = false;
    hangar.ftask = def_task;
    hangar.stask = recy_task;
    hangar.cnst = nil;

    local dropship_controller = mUnit.controller();
    dropship_controller:addListener("addObject","dropship_add");
    dropship_controller:addUnitListener("update","dropship_update");
    dropship_controller:addUnitListener("command_changed","dropship_command");
    dropship_controller:addUnitListener("receive","dropship_receive");
    
    dropship_controller:applyToODF("avdrop");
    
    local memTimer = mTimer.timer(2,true);
    memTimer:addListener("timeout","mem_timeout");
    --memTimer:start();
    
    intro_task:start();
    attack_task:start();
    def_task:start();
end


function Update(dtime)
    bzUtils:Update(dtime);
end


function GameKey(key)
    bzUtils:GameKey(key);
end

function AddObject(handle)
    bzUtils:AddObject(handle);
end

function CreateObject(handle)
    bzUtils:CreateObject(handle);
end

function DeleteObject(handle)
    bzUtils:DeleteObject(handle);
end

function Load(data)
    bzUtils:Load(data);
end

function Save()
    return bzUtils:Save();
end

init()


